


#include "cvutils.h"


namespace ltflo {


class TrackerState {
public:

	TrackerState(unsigned frameNo, const cv::Mat& image, const Gradient& gr, const cv::Mat& ptQuality, const cv::Mat& weightsField,
				BBox trackedArea, const std::vector<cv::Point>& border, const std::vector<double>& lines)
			: frameNo(frameNo), image(image.clone()), gr(gr.clone()), ptQuality(ptQuality.clone()), weightsField(weightsField.clone()),
			trackedArea(trackedArea), border(border), lines(lines), timesUsed(0.0) {}





	// getters
	std::vector<double>& getLines(void)		{ return lines; }
	// const getters
	unsigned 				getFrameNo(void)		const { return frameNo; }
	const cv::Mat&			getImage(void)			const { return image; }
	const Gradient&			getGradient(void)		const { return gr; }
	const cv::Mat&			getPtQuality(void)		const { return ptQuality; }
	const cv::Mat&			getWeightsField(void)	const { return weightsField; }
	const BBox&				getTrackedArea(void)	const { return trackedArea; }
	const std::vector<cv::Point>& getBorder(void)	const { return border; }
	const std::vector<double>& getLines(void)		const { return lines; }
	double					getTimesUsed(void)		const { return timesUsed; }

	
	// fade usage by a constant factor, such that it's local only
	void fadeUsage(void) {
		timesUsed *= modelUsageFadeFactor;
	}
	// and increase in case it was used
	void increaseUsage(void) {
		timesUsed += 1;
	}

private:
	// frame number from where this state was learnt
	unsigned frameNo;
	// image data and its gradient
	cv::Mat image;
	Gradient gr;
	// edge quality field and its blurred variant
	cv::Mat ptQuality, weightsField;
	// target location
	BBox trackedArea;
	// edge points
	std::vector<cv::Point> border;
	// lines (as a vector with abc coefficients)
	std::vector<double> lines;
	// how many times this model has been used for a correction?
	double timesUsed;














public:
	// fading factor for "timesUsed", as we care only locally
	static const double modelUsageFadeFactor = .995;





};















}




