/* Library for handling with correspondences, mostly 1D edge searches... */
#include "corrs.h"
//#include <mex.h>


namespace ltflo {

using std::min;
using std::max;

using namespace cv;

void unguidedEdgeSearch(unsigned *correspFull,
						const Gradient gr, double prefLen,
						unsigned LEN) {
	
	double maxLenMul = 3;
	double sigmaRatio = .5;
	double ww = prefLen * maxLenMul;
	unsigned ptsNo = 2*ceil(ww) + 1;
	double sigSq = ww * ww * sigmaRatio * sigmaRatio;
	
	unsigned i, j;
	unsigned px, py, qx, qy, mx, my;
	double xS, xD, yS, yD;
	double w, wMax, pa;
	
	unsigned M = gr.m.size().height;
	unsigned N = gr.m.size().width;

	for (i = 0; i < LEN; ++i) {
		mx = correspFull[2];
		my = correspFull[3];
		px = 0;
		py = 0;
		wMax = 0;
		
		while(px != mx || py != my) { /* Iterate each point until convergence */
			px = mx;
			py = my;
		
//			pa = gr.a[(px-1)*M + py - 1]; /* angle */
			pa = gr.a.at<double>(py, px); /* angle */
		
			xS = px - round(ww*cos(pa));
			xD = 2*round(ww*cos(pa)) / (ptsNo - 1);
			yS = py - round(ww*sin(pa));
			yD = 2*round(ww*sin(pa)) / (ptsNo - 1);

			for (j = 0; j < ptsNo; ++j) {
				qx = min(double(N-1), max(0., round(xS + j*xD)));
				qy = min(double(M-1), max(0., round(yS + j*yD)));
//				w = gr.m[(qx-1)*M + qy - 1];
				w = gr.m.at<double>(qy, qx);
				w *= exp( -( (px-(double)qx)*(px-(double)qx) + (py-(double)qy)*(py-(double)qy) ) / sigSq );
				if (wMax < w) {
					wMax = w;
					mx = qx;
					my = qy;
				}
			}
		}
		correspFull[2] = mx;
		correspFull[3] = my;
		correspFull += 4;
	}
}



void guidedEdgeSearchIteration(unsigned *correspFull,
								const Gradient grOld, const Gradient grNew, const cv::Mat& imOld, const cv::Mat& imNew,
								double rot, double prefLen, bool repair, unsigned LEN) {
	
	double dA = M_PI/10; /* side angle difference */
	int nA = 1; /* number of side angles on each side */
	
	unsigned maxLenMul = 3;
	double ww = prefLen*maxLenMul;
	unsigned ptsNo = 2*ceil(ww)+1;
	double endRatio = .75;
	
	unsigned i, j;
	int a;
	unsigned px, py, px2, py2;
	double pa, angle;
	unsigned char escaped;
	
	double *app1, *app2; /*Speed-up: some shared buffer?*/
	double xS, xE, xD, yS, yE, yD;
	int pxc = 0, pyc = 0, pxp, pyp, pxn = 0, pyn = 0; /* Candidate point, plus previous and next */
	double gmc = 0, gmp, gmn = 0; /* Gradient magnitude, plus previous and next */
	double w, wMax, dN, cosDist;
	unsigned M = imNew.size().height, N = imNew.size().width, CC = imNew.channels();
	
	app1 = (double *) malloc((2*ceil(prefLen) + 1) * CC * sizeof(double));
	app2 = (double *) malloc((2*ceil(prefLen) + 1) * CC * sizeof(double));
	
	for (i = 0; i < LEN; ++i) {
		wMax = 0;
		escaped = 0;
	
		px  = correspFull[0];
		py  = correspFull[1];
		px2 = correspFull[2];
		py2 = correspFull[3];
//		pa = grOld.a[(px-1)*M + py-1];
		pa = grOld.a.at<double>(py, px);
		loadAppearance(app1, imOld, px, py, pa, prefLen);
		
		/* Loop through 2*nA+1 angles symmetrically around central gradient */
		for(a = -nA, angle = pa - nA*dA + rot; a <= nA; ++a, angle += dA) {
			xS = px2 - round(ww*cos(angle));
			xE = px2 + round(ww*cos(angle)); /* speed-up: remove xE, round(ww*cos) directly to xD */
			xD = (xE - xS) / (ptsNo - 1);
			yS = py2 - round(ww*sin(angle));
			yE = py2 + round(ww*sin(angle));
			yD = (yE - yS) / (ptsNo - 1);
			
			for (j = 0; j < ptsNo; ++j) { /* candidate points */
				/* non-max sup */
				if (j == 0 || escaped) { /* Compute position and gradient of all */
					escaped = 0;
					pxp = (int)round(xS + (j-1)*xD);
					pyp = (int)round(yS + (j-1)*yD);
					if (pxp < 0 || pxp >= N || pyp < 0 || pyp >= M) {
						escaped = 1;
						continue;
					}
					gmp = grNew.m.at<double>(pyp, pxp); //[(pxp-1)*M + pyp - 1];
					pxc = (int)round(xS + j*xD);
					pyc = (int)round(yS + j*yD);
					if (pxc < 0 || pxc >= N || pyc < 0 || pyc >= M) {
						escaped = 1;
						continue;
					}
					gmc = grNew.m.at<double>(pyc, pxc);//[(pxc-1)*M + pyc - 1];
					pxn = (int)round(xS + (j+1)*xD);
					pyn = (int)round(yS + (j+1)*yD);
					if (pxn < 0 || pxn >= N || pyn < 0 || pyn >= M) {
						escaped = 1;
						continue;
					}
					gmn = grNew.m.at<double>(pyn, pxn);//[(pxn-1)*M + pyn - 1];
				} else { /* Use as much as possible and compute only the new ones */
					pxp = pxc;
					pyp = pyc;
					if (pxp < 0 || pxp >= N || pyp < 0 || pyp >= M) {
						escaped = 1;
						continue;
					}
					gmp = gmc;
					pxc = pxn;
					pyc = pyn;
					if (pxc < 0 || pxc >= N || pyc < 0 || pyc >= M) {
						escaped = 1;
						continue;
					}
					gmc = gmn;
					pxn = (int)round(xS + (j+1)*xD);
					pyn = (int)round(yS + (j+1)*yD);
					if (pxn < 0 || pxn >= N || pyn < 0 || pyn >= M) {
						escaped = 1;
						continue;
					}
					gmn = grNew.m.at<double>(pyn, pxn);//[(pxn-1)*M + pyn - 1];
				}
				if (gmc < gmp || gmc < gmn) { /* discard also points with very low magnitude of gradient? */
					continue;
				}
				/* /non-max sup */
				
				w = 1;
				
				/* 1) Distance: */
				dN = sqrt((px2 - pxc)*(px2 - pxc) + (py2 - pyc)*(py2 - pyc)) / ww;
				w *= min(1., 1. + (1.-endRatio) / 2. - 3./2. * (1.-endRatio) * dN);
				if (w < wMax) { /* speed-up */
/*					printf("I'm outta here, wheeee...\n");*/
					continue;
				}
				
				/* 2) Gradient */
				cosDist = cos(pa + rot - grNew.a.at<double>(pyc, pxc));//[(pxc-1)*M + pyc - 1]);
				if (repair) { /* Not oriented gradient */
					w *= fabs(cosDist);
				} else { /* Oriented gradient */
					w *= .5 + cosDist/2.;
				}
				if (w < wMax) { /* speed-up, app loading and comparing too expensive */
/*					printf("I'm outta here, wheeee...\n");*/
					continue;
				}
				
				/* 3) Appearance */
//				loadAppearance(app2, imNew, pxc, pyc, grNew.a[(pxc-1)*M + pyc - 1], prefLen, M, N, CC);
				loadAppearance(app2, imNew, pxc, pyc, grNew.a.at<double>(pyc,pxc), prefLen);
				w *= appearanceSimilarity(app1, app2, (unsigned)ceil(prefLen), CC);
				
				/* Check for the best so far */
				if (wMax < w) {
					wMax = w;
					correspFull[2] = pxc;
					correspFull[3] = pyc;
//					printf("Moving [%u;%u]->[%u,%u]...\n", px2, py2, pxc, pyc);
				}
			}
		}
		
		correspFull += 4;
	}
	
	free(app1);
	free(app2);
}



double appearanceSimilarity(double *a1, double *a2, unsigned halfLen, unsigned CC) {
	unsigned centre = halfLen + 1, shift = CC*halfLen;
	double ssd1 = 0, ssd2 = 0;
	unsigned i;
	
	for(i = 0; i < CC*centre; ++i) {
		ssd1 += (a1[i] - a2[i]) * (a1[i] - a2[i]);
		ssd2 += (a1[i+shift] - a2[i+shift]) * (a1[i+shift] - a2[i+shift]);
	}

	return (1 - 5.*sqrt(min(ssd1, ssd2)/((double)CC*centre))); /* SQRT?!? */
}



void loadAppearance(double *app, const cv::Mat& im, unsigned px, unsigned py, double pa, double len) {
	double xS, xE, xD, yS, yE, yD;
	unsigned ptsNo = 2*ceil(len)+1;
	unsigned i, c, li;
	unsigned M = im.size().height, N = im.size().width, CC = im.channels();
	
	xS = px - round(len*cos(pa));
	xE = px + round(len*cos(pa));
	xD = (xE - xS) / (ptsNo - 1);
	yS = py - round(len*sin(pa));
	yE = py + round(len*sin(pa));
	yD = (yE - yS) / (ptsNo - 1);
	
	for (i = 0; i < ptsNo; ++i) {
		px = (unsigned)max(0., min(double(N-1), round(xS + i*xD)));
		py = (unsigned)max(0., min(double(M-1), round(yS + i*yD)));
//		li = (px-1)*M + py - 1; /* linear index from coordinates */
		li = px + py*N; // row-major storage in OpenCV's Mat
		for (c = 0; c < CC; ++c) {
			app[i*CC + c] = im.data[li*CC + c] / 255.0;
			
		}
	}
	
}


unsigned extractLines(std::vector<double>& lines, const std::vector<cv::Point>& edgePoints, const Gradient& gr, double pointShift) {
	lines.resize(3 * edgePoints.size());
	std::vector<unsigned> edgePs(2 * edgePoints.size());
	for (unsigned point = 0; point < edgePoints.size(); ++point) {
		edgePs[point*2] = edgePoints[point].x;
		edgePs[point*2+1] = edgePoints[point].y;
	}
	return getLines(&lines[0], &edgePs[0], gr, edgePoints.size(), pointShift);
}

unsigned extractLines(std::vector<double>& lines, const std::vector<unsigned>& correspFull, const Gradient& gr, double pointShift) {
	if (correspFull.size() % 4) {
		throw "We didn't get the proper 4*N vector!";
	}
	unsigned LEN = correspFull.size() / 4;
	lines.resize(3 * LEN);
	std::vector<unsigned> edgePs(2 * LEN);
	for (unsigned point = 0; point < LEN; ++point) {
		edgePs[point*2] = correspFull[point*4+2];
		edgePs[point*2+1] = correspFull[point*4+3];
	}
	return getLines(&lines[0], &edgePs[0], gr, LEN, pointShift);
}



} // namespace ltflo


