#ifndef __GEOMETRY_H__
#define __GEOMETRY_H__

#include <cstdlib>
#include <cmath>
#include <vector>
#include "algebra.h"
#include "intersection.h"
#include "cvutils.h"
//#include <opencv2/core/core.hpp>

namespace ltflo {


#define xMin (limits[0])
#define yMin (limits[1])
#define xMax (limits[2])
#define yMax (limits[3])

typedef struct { /* Array to be passed as an additional parameter to NM and other optimisations. */
	const double * l1, * l2, * win;
	unsigned size;
	double *buffer, *d;
} arr6;

typedef struct { /* A similar array for A refine using weightsField */
	const double *wf, *u2;
	unsigned M, N, LEN;
} arr5;

typedef struct { /* A similar array for A refine using lines-point distance error */
	const double * l1, * l2, * u1, * u2;
	unsigned size;
	double * buffer;
} arr4D;

typedef struct { /* In both cartesian and polar coordinates (bitmaps) */
	double *x, *y, *a, *m;
} GradientD;

// cost function for A refinement, similar to arr5
class WeightsF:public cv::MinProblemSolver::Function{
public:
	int getDims() const { return 4; } // 4 for similarity
	double calc(const double* p) const;
	WeightsF(const cv::Mat& wf, const cv::Mat& u2) : wf(wf), u2(u2) {}
private:
	const cv::Mat& wf;
	const cv::Mat& u2;
};


////////////////////// SIMON'S STUFF, NOT NEEDED FOR NOW! ///////////////////////////


///* Functor for Simon's magic: line correspondences geom. error */
//class LineErrorF : fw::Opt_func {
//	public:
//		LineErrorF(const double * l1, const double * l2, const double * u1, const double * u2, unsigned size, double * buffer) :
//					l1(l1), l2(l2), u1(u1), u2(u2), size(size), buffer(buffer), buffer2(buffer), ownBuffer(false) {}; // here the buffer must be at least 12*size

//		LineErrorF(const double * u, unsigned size, const unsigned * inliers, unsigned uLen, double * inbuffer) : size(size) { // here the buffer must be at least 24*size
//			const double * ui;
//			double * l1nc, * l2nc, * u1nc, *u2nc;
//			if ( inbuffer ) {
//				this->buffer = inbuffer;
//				ownBuffer = false;
//			} else {
//				this->buffer = new double[24*size];
//				ownBuffer = true;
//			}
//			l1 = this->buffer + 0 * size;
//			l2 = this->buffer + 3 * size;
//			u1 = this->buffer + 6 * size;
//			u2 = this->buffer + 9 * size;
//			l1nc = (double *)l1; l2nc = (double *)l2; u1nc = (double *)u1; u2nc = (double *)u2; // pointer cast is evil :)
//			buffer2 = this->buffer + 12 * size;
//			
//			for ( unsigned i = 0; i < size; ++i ) {
//			// u is in form in form [l1;l2;x1;x2;ind]
//				if ( inliers ) {
//					ui = u + inliers[i] * uLen;
//				} else {
//					ui = u + i * uLen;
//				}
//				*l1nc++ = *ui++;
//				*l1nc++ = *ui++;
//				*l1nc++ = *ui++;
//				*l2nc++ = *ui++;
//				*l2nc++ = *ui++;
//				*l2nc++ = *ui++;
//				*u1nc++ = *ui++;
//				*u1nc++ = *ui++;
//				*u1nc++ = *ui++;
//				*u2nc++ = *ui++;
//				*u2nc++ = *ui++;
//				*u2nc++ = *ui++;
//			}
//		};
//		
//		std::vector<double> operator()(const std::vector<double> & cam_pos) const;
//		~LineErrorF() {
//			if ( ownBuffer ) {
//				delete[] buffer;
//			}
//		}
//		
//	private:
//		LineErrorF(){};
//		LineErrorF(const LineErrorF & other){};
//		LineErrorF & operator=(const LineErrorF & other){return *this;};
//		
//		const double * l1, * l2, * u1, * u2;
//		unsigned size;
//		double * buffer, * buffer2;
//		bool ownBuffer;
//};

///* Functor for Simon's magic: line correspondences geom. error (the original geom. error as in ACCV DTCE 2012) */
//class LineErrorOrigF : fw::Opt_func {
//	public:
//		LineErrorOrigF(const double * l1, const double * l2, const double * win, unsigned size, double * buffer) :
//					l1(l1), l2(l2), win(win), size(size), buffer(buffer), buffer2(buffer), ownBuffer(false) {}; // here the buffer must be at least 6*size

//		LineErrorOrigF(const double * u, const double * win, unsigned size, const unsigned * inliers, unsigned uLen, double * inbuffer) :
//					win(win), size(size) { // here the buffer must be at least 12*size
//			const double * ui;
//			double * l1nc, * l2nc;
//			if ( inbuffer ) {
//				this->buffer = inbuffer;
//				ownBuffer = false;
//			} else {
//				this->buffer = new double[12*size];
//				ownBuffer = true;
//			}
//			l1 = this->buffer + 0 * size;
//			l2 = this->buffer + 3 * size;
//			l1nc = (double *)l1; l2nc = (double *)l2; // pointer cast is evil :)
//			buffer2 = this->buffer + 6 * size;
//			
//			for ( unsigned i = 0; i < size; ++i ) {
//			// u is in form in form [l1;l2;x1;x2;ind]
//				if ( inliers ) {
//					ui = u + inliers[i] * uLen;
//				} else {
//					ui = u + i * uLen;
//				}
//				*l1nc++ = *ui++;
//				*l1nc++ = *ui++;
//				*l1nc++ = *ui++;
//				*l2nc++ = *ui++;
//				*l2nc++ = *ui++;
//				*l2nc++ = *ui++;
//			}
//		};
//		
//		std::vector<double> operator()(const std::vector<double> & cam_pos) const;
//		~LineErrorOrigF() {
//			if ( ownBuffer ) {
//				delete[] buffer;
//			}
//		}
//		
//		inline const double * getL1() {return l1;}
//		inline const double * getL2() {return l2;}
//		
//		
//	private:
//		LineErrorOrigF(){};
//		LineErrorOrigF(const LineErrorOrigF & other){};
//		LineErrorOrigF & operator=(const LineErrorOrigF & other){return *this;};
//		
//		const double * l1, * l2, * win;
//		unsigned size;
//		double * buffer, * buffer2;
//		bool ownBuffer;
//};

///* Functor for Simon's magic: gradient of line correspondences geom. error (the original geom. error as in ACCV DTCE 2012) */
//class LineErrorOrigG : fw::Grad_func {
//	public:
//		LineErrorOrigG(const double * l1, const double * l2, const double * win, unsigned size, fw::Opt_func * base_f) :
//					fw::Grad_func(base_f), l1(l1), l2(l2), win(win), size(size) {}; // here the buffer is not necessary

//		int operator()(const std::vector<double> & cam_pos, cv::Mat & gradients, int basePtr) const;

//	private:
//		LineErrorOrigG(){};
//		LineErrorOrigG(const LineErrorOrigG & other){};
//		LineErrorOrigG & operator=(const LineErrorOrigG & other){return *this;};
//		
//		const double * l1, * l2, * win;
//		unsigned size;
////		double * buffer, * buffer2;
////		bool ownBuffer;
//};

///* Functor for Simon's magic: line correspondences geom. error */
//class EvidF : fw::Opt_func {
//	public:
//		EvidF (const double * u, unsigned I, const unsigned * inliers, unsigned u_len, const double * evidDist, const double * evidAng, unsigned M, unsigned N, bool repair)
//				: u(u), evidDist(evidDist), evidAng(evidAng), size(I), M(M), N(N), U_SIZE(u_len), inliers(inliers), repair(repair) {};
//		
//		std::vector<double> operator()(const std::vector<double> & cam_pos) const;
//		
//	private:
//		EvidF(){};
//		EvidF(const EvidF & other){};
//		EvidF & operator=(const EvidF & other){return *this;};
//		
//		const double * u, * evidDist, * evidAng;
//		unsigned size, M, N, U_SIZE;
//		const unsigned * inliers;
//		bool repair;
//};

/*function [d] = lineDistSq(l1, l2, win) % lines have to be already normalized!*/
void lineDistSq(double * d, const double * l1, const double * l2, unsigned len, const double * win);

/*floating point modulo a%b */
inline double doubleMod(double a, double b) {
	return a - floor(a/b)*b;
}

/*function [d] = aLineDistSq(a, l1, l2, win)*/
/* buffer has to be sized 7*len */
void aLineDistSq(double * d, const double * A, const double * l1, const double * l2, unsigned len, const double * win, double * buffer);

/* Separated components of the original geometric error
	Lines have to be already normalized! */
void lineDistSeparated(double * d, const double * l1, const double * l2, unsigned len, const double * win);

/*function [d] = aError(a, u(:,inliers))
	returned 'd' is of size 4xLEN!
	buffer has to be sized 24*LEN
	pass NULL in 'inliers' if all the points are measured */
void aErrorU(double * d, const double * A, const double * u, const unsigned * inliers, unsigned len, unsigned uLen, double * buffer);
/* The same as above, but d is sum of squared individual errors */
void aErrorSqU(double * d, const double * A, const double * u, const unsigned * inliers, unsigned len, unsigned uLen, double * buffer);
/*function [d] = aError(a, u(:,inliers))
	returned 'd' is of size 4xLEN!
	buffer has to be sized 12*LEN */
void aError(double * d, const double * A, const double * l1, const double * l2, const double * u1, const double * u2, unsigned len, double * buffer);

/*l2 = inv(a^T) l1*/
void aTInvMultiply(double * l2, const double * a, const double * l1, unsigned len);
/*l2 = inv(a) l1*/
void aInvMultiply(double * l2, const double * a, const double * l1, unsigned len);
/*l2 = a^T l1*/
void aTMultiply(double * l2, const double * a, const double * l1, unsigned len);
/*l2 = a l1*/
void aMultiply(double * l2, const double * a, const double * l1, unsigned len);

/**
 * projectA
 *
 * @param double *xp         O  LEN   x-coord. of projected points
 * @param double *yp         O  LEN   y-coord. of projected points
 * @param const double *A    I  3x3   affine (or simpler) transformation, NOT HOMOGRAPHY!
 * @param const unsigned *x  I  LEN   x-coord of points to project
 * @param const unsigned *y  I  LEN   y-coord of points to project
 * @param unsigned LEN       I scalar number of points to project
 *
 * @brief Projects LEN points [x;y] by affine transformation A.
 */
void projectA(double *xp, double *yp, const double *A, const unsigned *x, const unsigned *y, unsigned LEN);
// projects one cv::Point or something similar
template <typename Pt>
Pt projectPt(const cv::Mat& A, const Pt& p);
// projects a vector of points or something similar, such as BBox
template <typename VecOfPts>
VecOfPts projectPts(const cv::Mat& A, const VecOfPts& pts);
// this one does the projection and additionally creates the "correspFull" structure with original points as well and rounds the result
std::vector<unsigned> projectCorrespondences(const cv::Mat& A, const std::vector<cv::Point>& points, cv::Size limits = cv::Size(1e6, 1e6));


/**
 * projectInvA
 *
 * @param double *xp         O  LEN   x-coord. of projected points
 * @param double *yp         O  LEN   y-coord. of projected points
 * @param const double *A    I  3x3   affine (or simpler) transformation, NOT HOMOGRAPHY!
 * @param const unsigned *x  I  LEN   x-coord of points to project
 * @param const unsigned *y  I  LEN   x-coord of points to project
 * @param unsigned LEN       I scalar number of points to project
 *
 * @brief Backprojects LEN points [x;y] by inverse of affine transformation A.
 */
void projectInvA(double *xp, double *yp, const double *A, const unsigned *x, const unsigned *y, unsigned LEN);

/*Normalize to unitary length on normal vector*/
void normLines(double * l, unsigned len);
/*Normalize to unitary third coordinate*/
void normPoints(double * u, unsigned len);

/*aLineDistSq with compact correspondence input, just wrapper*/
/* buffer have to be sized 13*len */
void aLineDistSqU(double * d, const double * A, const double * u, unsigned len, const double *win, unsigned correspLength, double *buffer);

/*Area of a rectangle*/
double rectangleArea(point * rect);

/*computes sum of squared geometric errors of line correspondences*/
void geomError(double *p, double *hx, int m, int n, void *adata);

/*computes sum of squared geometric errors of line correspondences - distance-angle measure*/
double geomErrorSimple(double *p, void *adata);

/*computes sum of squared geometric errors of line correspondences - four line-point distances*/
double geomErrorSimple4D(double *p, void *adata);

/* Computes negative sum of weights at backprojected positions */
double refineFitError(double *p, void *adata);

/* function [lines] = getLines(border, gr) */
/* Get lines by position and gradient orientation */
/*
Arguments:
lines      O  3xLEN   homogeneous coordinates of lines fitted to the gradient orientation at given positions
corresp    I  2xLEN   correspondences
gr         I  4(MxN)  gradient (m,a,x,y)
M, N, LEN  I  scalar  image dimensions, number of correspondences
pointShift I  scalar  an offset (in px), added to all point coordinates when creating lines (set to 1 for 1-based lines from 0-based points)
Returns:
Number of orientation reversions
*/
unsigned getLines(double *lines, const unsigned *corresp, const Gradient gr, unsigned LEN, double pointShift = 0);

double fastTan1 (double fAngle);
double fastInvTan1 (double fValue);
double lutTan(double angle);
double lutTanSimple(double angle);
double lutATan(double angle);
double lutATanS(double angle);
double lutInterpTan(double angle);
double lutInterpATan(double angle);
double lutInterpATanS(double value);
extern double TAN_TABLE[];
/* Simple - we know the table is set and the angle is > 0 and < PI/2 */
inline double lutInterpTanS(double angle) {
	double index;
	unsigned indexH, indexL;
	
	index = angle*1000;
	indexL = (unsigned)index;
	indexH = indexL + 1;
	return TAN_TABLE[indexL]*(indexH-index) + TAN_TABLE[indexH]*(index-indexL);
}
extern double ATAN_TABLE[];
/* Simple - we know the table is set */
inline double lutInterpATanS(double value) {
	double index;
	unsigned indexH, indexL;
	
	if (value < 0) {
		return -lutInterpATanS(-value); /* Odd function */
	}
	if (value >= 1000) {
		return M_PI / 2;
	} /* Now the value is in range [0, pi/2] */
	
	if (value < 10) {
		index = value*1000;
	} else if (value < 100) {
		index = 10000 + value*10 - 100;
	} else {
		index = 10900 + value*0.1 - 10;
	}

	indexL = (unsigned)index;
	indexH = indexL + 1;
	return ATAN_TABLE[indexL]*(indexH-index) + ATAN_TABLE[indexH]*(index-indexL);
}


/* function [ptQuality] = updatePtQuality(ptQualityOld, correspFull, lines, A, ptQualityDec, evid) */
/* Update point quality by warping and adding ones at point positions */
/*
Arguments:
ptQualNew    O   MxN   transformed field of point qualities
ptQualOld    I   MxN   field of point qualities in the previous frame f_t-1
corresp      I  4xLEN  correspondences from f_t-1 to f_t
lines        I  3xLEN  lines in the previous frame f_t-1, homogeneous coordinates
A            I   3x3   transformation from f_t-1 to f_t
ptQualityDec I  scalar fading factor
evidDist     I   MxN   chamfer distance field in f_t
evidAng      I   MxN   angle of closest edge point in f_t
M, N, LEN    I  scalar image dimensions, number of correspondences
*/
void updatePtQuality(double *ptQualNew, const double *ptQualOld, const unsigned *corresp, const double *lines, const double *A,
					double ptQualityDec, const double *evidDist, const double *evidAng, unsigned M, unsigned N, unsigned LEN);

// a wrapper for the above
void updatePtQualityWrapper(cv::Mat& ptQuality,
					const std::vector<unsigned> corresp, const std::vector<double> lines, const cv::Mat& A, double ptQualityDec, const Evidence evidence);


/**
 * getGradient
 *
 * @param Gradient gr       O 4(MxN)  structure with gradient
 * @param const double *img I  MxN    input image, grayscale
 * @param unsigned M        I scalar  image height
 * @param unsigned N        I scalar  image width
 * @param unsigned *limits  I   4     computation limits (xMin, yMin, xMax, yMax)
 *
 * @brief Computes gradient of a gray image -- components, angle and magnitude.
 */
void getGradient(GradientD gr, const double *img, unsigned M, unsigned N, unsigned *limits);




















// projects one cv::Point or something similar
template <typename Pt>
Pt projectPt(const cv::Mat& A, const Pt& p) {
	Pt res;
	res.x = A.at<double>(0,0) * p.x + A.at<double>(0,1) * p.y + A.at<double>(0,2);
	res.y = A.at<double>(1,0) * p.x + A.at<double>(1,1) * p.y + A.at<double>(1,2);
	return res;
}

// projects a vector of points or something similar
template <typename VecOfPts>
VecOfPts projectPts(const cv::Mat& A, const VecOfPts& pts) {
	VecOfPts res = pts; // just to copy its size etc...
	for (unsigned i = 0; i < pts.size(); ++i) {
		res[i] = projectPt(A, pts[i]);
	}
	return res;
}




} // namespace ltflo


#endif /* __GEOMETRY_H__ */

