#include "intersection.h"

namespace ltflo {

double intersectionArea(const BBox& ba, const BBox& bb) {
	point pA[4], pB[4];
	for (unsigned i = 0; i < ba.size(); ++i) {
		pA[i].x = ba[i].x;
		pA[i].y = ba[i].y;
		pB[i].x = bb[i].x;
		pB[i].y = bb[i].y;
	}
	return intersectionArea(4, pA, 4, pB);
}


/*Intersection of two polygons, from http://www.cap-lore.com/MathPhys/IP/*/
double intersectionArea(int na, point a[], int nb, point b[]){
	vertex *ipa, *ipb;
	double ascale;
	box B = {{bigReal, bigReal}, {-bigReal, -bigReal}};
	const double gamut = 500000000., mid = gamut/2.;
	double rngx, rngy, sclx, scly;
	long s = 0, a1, a2, a3, a4;
	int j, k, o;

	if(na < 3 || nb < 3) {
		return 0;
	}

	ipa = (vertex *) malloc((na+1)*sizeof(vertex));
	ipb = (vertex *) malloc((nb+1)*sizeof(vertex));

	range(&B, a, na);
	range(&B, b, nb);

	rngx = B.max.x - B.min.x;
	sclx = gamut/rngx;
	rngy = B.max.y - B.min.y;
	scly = gamut/rngy;

	fit(a, na, ipa, 0, sclx, scly, mid, B);
	fit(b, nb, ipb, 2, sclx, scly, mid, B);
	ascale = sclx*scly;

	for (j = 0; j<na; ++j) {
		for (k = 0; k<nb; ++k) {
			if (ovl(ipa[j].rx, ipb[k].rx) && ovl(ipa[j].ry, ipb[k].ry)) {
				a1 = -area(ipa[j].ip, ipb[k].ip, ipb[k+1].ip);
				a2 = area(ipa[j+1].ip, ipb[k].ip, ipb[k+1].ip);
				o = a1<0;
				if ( o == (a2<0) ) {
					a3 = area(ipb[k].ip, ipa[j].ip, ipa[j+1].ip);
					a4 = -area(ipb[k+1].ip, ipa[j].ip, ipa[j+1].ip);
					if ((a3<0) == (a4<0)) {
						if (o) {
							int_cross(&s, &ipa[j], &ipa[j+1], &ipb[k], &ipb[k+1], a1, a2, a3, a4);
						} else {
							int_cross(&s, &ipb[k], &ipb[k+1], &ipa[j], &ipa[j+1], a3, a4, a1, a2);
						}
					}
				}
			}
		}
	}

	inness(&s, ipa, na, ipb, nb);
	inness(&s, ipb, nb, ipa, na);

	free(ipa);
	free(ipb);
	return s/ascale;
}

void range(box *B, point * x, int c) {
	while (c--) {
		bd(&B->min.x, x[c].x);
		bu(&B->max.x, x[c].x);
		bd(&B->min.y, x[c].y);
		bu(&B->max.y, x[c].y);
	}
}

void fit(point * x, int cx, vertex * ix, int fudge, double sclx, double scly, double mid, box B){
	int c = cx;
	
	while(c--){
		ix[c].ip.x = (((long)((x[c].x - B.min.x)*sclx - mid)) & (~7)) | fudge | (c&1);
		ix[c].ip.y = (((long)((x[c].y - B.min.y)*scly - mid)) & (~7)) | fudge;
	}
	ix[0].ip.y += cx&1;
	ix[cx] = ix[0];
	
	c=cx;
	while(c--) {
		if (ix[c].ip.x < ix[c+1].ip.x) {
			ix[c].rx.mn = ix[c].ip.x;
			ix[c].rx.mx = ix[c+1].ip.x;
		} else {
			ix[c].rx.mn = ix[c+1].ip.x;
			ix[c].rx.mx = ix[c].ip.x;
		}
		if (ix[c].ip.y < ix[c+1].ip.y) {
			ix[c].ry.mn = ix[c].ip.y;
			ix[c].ry.mx = ix[c+1].ip.y;
		} else {
			ix[c].ry.mn = ix[c+1].ip.y;
			ix[c].ry.mx = ix[c].ip.y;
		}
		ix[c].in=0;
	}
}

long area(ipoint a, ipoint p, ipoint q) {
	return p.x*q.y - p.y*q.x + a.x*(p.y - q.y) + a.y*(q.x - p.x);
}


int ovl(rng p, rng q){
	return p.mn < q.mx && q.mn < p.mx;
}

void cntrib(long *s, ipoint f, ipoint t, long w) {
	*s += w*(t.x-f.x)*(t.y+f.y)/2;
}

void int_cross(long *s, vertex * a, vertex * b, vertex * c, vertex * d, double a1, double a2, double a3, double a4) {
	double r1=a1/(a1+a2), r2 = a3/(a3+a4);
	ipoint tempPoint;

	tempPoint.x = a->ip.x + r1*(b->ip.x - a->ip.x);
	tempPoint.y = a->ip.y + r1*(b->ip.y - a->ip.y);
	cntrib(s, tempPoint, b->ip, 1);
	tempPoint.x = c->ip.x + r2*(d->ip.x - c->ip.x);
	tempPoint.y = c->ip.y + r2*(d->ip.y - c->ip.y);
	cntrib(s, d->ip, tempPoint, 1);
	++a->in;
	--c->in;
}

void inness(long *s, vertex * P, int cP, vertex * Q, int cQ) {
	int ss=0, c=cQ, sgn, j;
	ipoint p = P[0].ip;
	while(c--) {
		if(Q[c].rx.mn < p.x && p.x < Q[c].rx.mx) {
			sgn = 0 < area(p, Q[c].ip, Q[c+1].ip);
			ss += ((sgn != (Q[c].ip.x < Q[c+1].ip.x)) ? 0 : (sgn?-1:1));
		}
	}
	for(j = 0; j<cP; ++j) {
		if(ss) {
			cntrib(s, P[j].ip, P[j+1].ip, ss);
		}
		ss += P[j].in;
	}
}

void bd(double * X, double y) {
	*X = *X<y ? *X : y;
}

void bu(double * X, double y) {
	*X = *X>y ? *X : y;
}


} // namespace ltflo


