#include <iostream>
#include <fstream>

#include "opencv2/opencv.hpp"

//#define VOT_RECTANGLE
#define VOT_POLYGON
#include "vot.h"
#include "LTFLO.h"

//#define DISPLAY

using namespace ltflo;

enum ExitCode { // error codes
	EXITCODE_SUCCESS = 0,
	EXITCODE_INPUT_ERR = 1,
	EXITCODE_OUTPUT_ERR = 2,
	EXITCODE_CALL_ERR = 3,
	
	
	EXITCODE_OTHER_ERR = -1
};

//----------------------------converters to/from VOT format
//convert a rectangular region that may be rotated to polygon
VOTRegion bbox2region(const BBox& b)
{
    
    VOTRegion out(4);
    out.set(0, b.p0.x, b.p0.y);
    out.set(1, b.p1.x, b.p1.y);
    out.set(2, b.p2.x, b.p2.y);
    out.set(3, b.p3.x, b.p3.y);
    return out;
}

//a polygon to a rectangular and axis aligned bbox
BBox region2AAbbox(const VOTRegion& r)
{
    float left=9999,right=-9999,top=9999,bottom=-99999;
    for(int i=0;i<r.count();i++)
    {
        left=min(left, r.get_x(i));
        right=max(right, r.get_x(i));
        top=min(top, r.get_y(i));
        bottom=max(bottom, r.get_y(i));
    }
    
    return BBox::BBoxLTRB(left, top, right, bottom);
}

BBox region2bbox(const VOTRegion& r)
{
	cv::Point2d p1,p2,p3,p4;
	p1.x=r.get_x(3);	p1.y=r.get_y(3);
	p2.x=r.get_x(2);	p2.y=r.get_y(2);
	p3.x=r.get_x(1);	p3.y=r.get_y(1);
	p4.x=r.get_x(0);	p4.y=r.get_y(0);
	
    return BBox(p1,p2,p3,p4);
}

//-------------------------------------------------



//main taken from VOT ncc.cpp example
int main( int argc, char** argv)
{
    std::cout << "@@LTFLO launched"<<std::endl;

    // *****************************************
    // VOT: Call vot_initialize at the beginning
    // *****************************************
    VOT vot;
#ifdef DISPLAY
	cv::namedWindow("LT-FLO", cv::WINDOW_AUTOSIZE);
#endif

    // Process the first frame
    VOTRegion initialization = vot.region();
    cv::Mat image = cv::imread(vot.frame());
    
	LTFLO tracker(image, region2bbox(initialization));
	//tracker.getTrackedArea().print(image, cv::Scalar(0,0,255), 3);
	//videoOut.write(image);
#ifdef DISPLAY
    imshow("LT-FLO", image);
	cv::waitKey(100);
#endif

    BBox target,prev_target;
    prev_target=region2bbox(initialization);
    while (!vot.end())
    {
        // *****************************************
        // VOT: Call vot_frame to get path of the
        //      current image frame. If the result is
        //      null, the sequence is over.
        // *****************************************
        string imagepath = vot.frame();
        if (imagepath.empty()) break;
        image = cv::imread(imagepath);
        
        //Run tracker and display
		tracker.track(image);
		target = tracker.getTrackedArea();
		if (target.isInvalid()) {
			//return 0;
            		target = prev_target;
            
		    //rebuild tracker
	//	    std::cerr<<"Rebuilding tracker with target: "<<target<<std::endl;
		    tracker = LTFLO(image, target);
		}
        prev_target=target;

        // *****************************************
        // VOT: Report the position of the object
        //      every frame using vot_report function.
        // *****************************************
        vot.report(bbox2region(target));

    }

    return 0;
}








